
#include <math.h>
#include <malloc.h>
#include <minmax.h>
#include "resizer_yuy2.h"

void * aligned_malloc( int size, int align )
{
	//return( malloc( size ) );
	char *memory, *temp_mem;
	if ( memory = (char*)malloc( size+align+8 ) ) {
		temp_mem = memory;
		memory += 4+align-1;
		memory = memory - ((long)memory%align);
		((char**)memory)[-1] = temp_mem;
		return( (void*)memory );
	}
	return( 0 );
}

void aligned_free( void *memory )
{
	if ( memory ) {
		//free( memory );
		char *temp_mem;	
		temp_mem = ((char**)memory)[-1];
		free( temp_mem );
	}
}

class ResamplingFunction {
public:
  virtual double f(double x) = 0;
  virtual double support() = 0;
};

class MitchellNetravaliFilter : public ResamplingFunction {
  double p0,p2,p3,q0,q1,q2,q3;
public:
  MitchellNetravaliFilter(double b=1./3., double c=1./3.) {
    p0 = (  6. -  2.*b        ) / 6.;
    p2 = (-18. + 12.*b +  6.*c) / 6.;
    p3 = ( 12. -  9.*b -  6.*c) / 6.;
    q0 = (	     8.*b + 24.*c) / 6.;
    q1 = (	  - 12.*b - 48.*c) / 6.;
    q2 = (	     6.*b + 30.*c) / 6.;
    q3 = (     -     b -  6.*c) / 6.;
  }
  double f(double x) {
    x = fabs(x);
    return (x<1) ? (p0+x*x*(p2+x*p3)) : (x<2) ? (q0+x*(q1+x*(q2+x*q3))) : 0.0;
  }
  double support() { return 2.0; }
};

// This function returns a resampling "program" which is interpreted by the
// FilteredResize filters.  It handles edge conditions so FilteredResize
// doesn't have to.

// Original value: 65536
// 2 bits sacrificed because of 16 bit signed MMX multiplication
const int FPScale = 16384; // fixed point scaler

static int* GetResamplingPatternRGB(int original_width, double subrange_start, double subrange_width,
  int target_width, ResamplingFunction* func )
{
  double scale = double(target_width) / subrange_width;
  double filter_step = min(scale, 1.0);
  double filter_support = func->support() / filter_step;
  int fir_filter_size = int(ceil(filter_support*2));
  int* result = (int*)aligned_malloc((1 + target_width*(1+fir_filter_size)) * 4, 4);

  int* cur = result;
  *cur++ = fir_filter_size;

  double pos_step = subrange_width / target_width;
  // the following translates such that the image center remains fixed
  double pos = subrange_start + ((subrange_width - target_width) / (target_width*2));

  for (int i=0; i<target_width; ++i) {
    int end_pos = int(pos + filter_support);
    if (end_pos > original_width-1)
      end_pos = original_width-1;
    int start_pos = end_pos - fir_filter_size + 1;
    if (start_pos < 0)
      start_pos = 0;
    *cur++ = start_pos;
    // the following code ensures that the coefficients add to exactly FPScale
    double total = 0.0;
    for (int j=0; j<fir_filter_size; ++j)
      total += func->f((start_pos+j - pos) * filter_step);
    double total2 = 0.0;
    for (int k=0; k<fir_filter_size; ++k) {
      double total3 = total2 + func->f((start_pos+k - pos) * filter_step) / total;
      *cur++ = int(total3*FPScale+0.5) - int(total2*FPScale+0.5);
      total2 = total3;
    }
    pos += pos_step;
  }
  return result;
}

// This is the same as for RGB except for the special treatments taken
// to support the YUV-MMX code.

static int* GetResamplingPatternYUV(int original_width, double subrange_start, double subrange_width,
  int target_width, ResamplingFunction* func, bool luma, unsigned char *temp)
{
  double scale = double(target_width) / subrange_width;
  double filter_step = min(scale, 1.0);
  double filter_support = func->support() / filter_step;
  int fir_filter_size = int(ceil(filter_support*2));
  int fir_fs_mmx = (fir_filter_size / 2) +1;  // number of loops in MMX code
  int* result = luma ?
    (int*)aligned_malloc(2*4 + target_width*(1+fir_fs_mmx)*8, 8) :
    (int*)aligned_malloc(2*4 + target_width*(1+fir_filter_size)*8, 8);

  int* cur[2] = { result +2, result +3 };
  *result = luma ? fir_fs_mmx : fir_filter_size;

  double pos_step = subrange_width / target_width;
  // the following translates such that the image center remains fixed
  double pos = subrange_start + ((subrange_width - target_width) / (target_width*2));

  for (int i=0; i<target_width; ++i) {
    int end_pos = int(pos + filter_support);
    if (end_pos > original_width-1)
      end_pos = original_width-1;
    int start_pos = end_pos - fir_filter_size + 1;
    if (start_pos < 0)
      start_pos = 0;
    int ii = luma ? i&1 : 0;
    *(cur[ii]) = luma ?
      (int)(temp + (start_pos & -2) * 2) :
      (int)(temp + start_pos * 8);
    cur[ii] += 2;
    // the following code ensures that the coefficients add to exactly FPScale
    double total = 0.0;
    for (int j=0; j<fir_filter_size; ++j)
      total += func->f((start_pos+j - pos) * filter_step);
    double total2 = 0.0;
    int oldCoeff = 0;
    for (int k=0; k<fir_filter_size; ++k)
    {
      double total3 = total2 + func->f((start_pos+k - pos) * filter_step) / total;
      int coeff = int(total3*FPScale+0.5) - int(total2*FPScale+0.5);
      total2 = total3;
      if (luma)
      {
        if ((k + start_pos) & 1)
        {
          *(cur[ii]) = (coeff << 16) + (oldCoeff & 0xFFFF);
          cur[ii] += 2;
        }
        else
          oldCoeff = coeff;
      }
      else
      {
        *(cur[0]) = coeff;  cur[0] += 1;
        *(cur[0]) = coeff;  cur[0] += 1;
      }
    }
    if (luma)
    {
      if ((start_pos + fir_filter_size) & 1)
      {
        *(cur[ii]) = 0 + (oldCoeff & 0xFFFF);
        cur[ii] += 2;
      }
      else
      if ((fir_filter_size & 1) == 0)
      {
        *(cur[ii]) = 0;  cur[ii] += 2;
      }
    }
    pos += pos_step;
  }
  return result;
}

void LetterBox_YUY2( char *Address, int pitch, int height, int top, int bottom )
{
	static const __int64 x8010801080108010 = 0x8010801080108010;

	char	*bot_point;
	int	t_count, b_count;

	bot_point = Address + pitch*(height-bottom);
	t_count = top*pitch/8;
	b_count = bottom*pitch/8;

	__asm {
        movq        mm7,x8010801080108010

		mov			edi,Address
		mov			edx,t_count
		cmp			edx,0
		je			overt
loopt:	movq        [edi],mm7
		add			edi,8
		dec			edx
		jnz			loopt

overt:	mov			edi,bot_point
		mov			edx,b_count
		cmp			edx,0
		je			overb
loopb:	movq        [edi],mm7
		add			edi,8
		dec			edx
		jnz			loopb

overb:	emms
	}
}

unsigned char *H_tempY = 0, *H_tempUV = 0;
int *H_pattern_luma = 0, *H_pattern_chroma = 0;

void FilteredResizeH( YUY2_Frame *src, YUY2_Frame *dst, double subrange_left, double subrange_width, bool flip )
{
	int original_width;

	int dst_pitch;
	char *dst_data;
	if ( flip ) {
		dst_pitch = -dst->pitch;
		dst_data = dst->data+dst->pitch*(dst->height-1);
	} else {
		dst_pitch = dst->pitch;
		dst_data = dst->data;
	}

    original_width = src->width;

	const unsigned char* srcp = (const unsigned char*)src->data;
	unsigned char* dstp = (unsigned char*)dst_data;

	int fir_filter_size_luma = H_pattern_luma[0];
	int fir_filter_size_chroma = H_pattern_chroma[0];

	static const __int64 x0000000000FF00FF = 0x0000000000FF00FF;
	static const __int64 x00FF000000FF0000 = 0x00FF000000FF0000;
	static const __int64 FPround =           0x0000200000002000;  // 16384/2

	__asm {
        emms
        pxor        mm0, mm0
        movq        mm7, x0000000000FF00FF
        movq        mm6, FPround
        movq        mm5, x00FF000000FF0000
	}

      for (int y=0; y<src->height; ++y)
      {
        int* cur_luma = H_pattern_luma+2;
        int* cur_chroma = H_pattern_chroma+2;
        int x = dst->width/2;//src->width/2;
        __asm {
          mov         ecx, [original_width]
          mov         edx, [H_tempY]
          mov         ebx, [H_tempUV]
          mov         esi, srcp
          mov         eax, -1
        // deinterleave current line
        deintloop:
          inc         eax
          movd        mm1, [esi]          ;mm1 = 0000VYUY
          movq        mm2, mm1
          punpcklbw   mm2, mm0            ;mm2 = 0V0Y0U0Y
          pand        mm1, mm7            ;mm1 = 00000Y0Y
          psrld       mm2, 16             ;mm2 = 000V000U
          movd        [edx+eax*4], mm1
          add         esi, 4
          movq        [ebx+eax*8], mm2
          sub         ecx, 2
          jnz         deintloop
        // use this as source from now on
          mov         eax, cur_luma
          mov         ebx, cur_chroma
          mov         edx, dstp
        xloopYUV:
          mov         esi, [eax]          ;esi=&tempY[ofs0]
          movq        mm1, mm0
          mov         edi, [eax+4]        ;edi=&tempY[ofs1]
          movq        mm3, mm0
          mov         ecx, fir_filter_size_luma
		  add         eax, 8              ;cur_luma++
        aloopY:
          // Identifiers:
          // Ya, Yb: Y values in srcp[ofs0]
          // Ym, Yn: Y values in srcp[ofs1]
          movd        mm2, [esi]          ;mm2 =  0| 0|Yb|Ya
          punpckldq   mm2, [edi]          ;mm2 = Yn|Ym|Yb|Ya
                                          ;[eax] = COn|COm|COb|COa
          pmaddwd     mm2, [eax]          ;mm2 = Y1|Y0 (DWORDs)
          paddd       mm1, mm2            ;accumulate
          add         esi, 4
          add         edi, 4
          add         eax, 8              ;cur_luma++
          dec         ecx
          jnz         aloopY

          mov         esi, [ebx]          ;esi=&tempUV[ofs]
          add         ebx, 8              ;cur_chroma++
          mov         ecx, fir_filter_size_chroma
        aloopUV:
          movq        mm2, [esi]          ;mm2 = 0|V|0|U
                                          ;[ebx] = 0|COv|0|COu
          pmaddwd     mm2, [ebx]          ;mm2 = V|U (DWORDs)
          paddd       mm3, mm2            ;accumulate
          add         esi, 8
          add         ebx, 8              ;cur_chroma++
          dec         ecx
          jnz         aloopUV

          paddd       mm1, mm6            ;Y1|Y1|Y0|Y0
          paddd       mm3, mm6            ; V| V| U| U
          pslld       mm3, 2
          pand        mm3, mm5            ;mm3 = v| 0|u| 0
          psrld       mm1, 14             ;mm1 = 0|y1|0|y0
          por         mm3, mm1
          packuswb    mm3, mm3            ;mm3 = ...|v|y1|u|y0
          movd        [edx], mm3
          add         edx, 4
          dec         x
          jnz         xloopYUV
        }
        srcp += src->pitch;
        dstp += dst_pitch;
      }
      __asm { emms }
};

int *V_resampling_pattern = 0;
int *V_yOfs = 0;

void FilteredResizeV( YUY2_Frame *src, YUY2_Frame *dst, double subrange_top, double subrange_height, bool flip )
{

	int dst_pitch;
	char *dst_data;
	if ( flip ) {
		dst_pitch = -dst->pitch;
		dst_data = dst->data+dst->pitch*(dst->height-1);
	} else {
		dst_pitch = dst->pitch;
		dst_data = dst->data;
	}

    const int* cur = V_resampling_pattern;
    const int fir_filter_size = *cur++;
    const int src_pitch = src->pitch;
    const int xloops = src->rowsize/4;
    const unsigned char* srcp = (const unsigned char*)src->data;
    unsigned char* dstp = (unsigned char*)dst_data;
    int y = dst->height;

	__asm {
		emms
		mov         edx, cur
      pxor        mm0, mm0
      mov         edi, fir_filter_size
    yloop:
		mov         esi, [V_yOfs]
      mov         eax, [edx]              ;eax = *cur
      mov         esi, [esi+eax*4]
      add         esi, srcp               ;esi = srcp + yOfs[*cur]
      add         edx, 4                  ;cur++
      xor         ecx, ecx                ;ecx = x = 0
    xloop:
      pxor        mm7, mm7
      pxor        mm1, mm1                ;total = 0
      lea         eax, [esi+ecx*4]        ;eax = srcp2 = srcp + x
      xor         ebx, ebx                ;ebx = b = 0
    bloop:
      movd        mm2, [eax]              ;mm2 = *srcp2 = 0|0|0|0|d|c|b|a
      movd        mm3, [edx+ebx*4]        ;mm3 = cur[b] = 0|co
      punpckldq   mm3, mm3                ;mm3 = co|co
      punpcklbw   mm2, mm0                ;mm2 = 0d|0c|0b|0a
      movq        mm4, mm2
      punpcklwd   mm2, mm0                ;mm2 = 00|0b|00|0a
      punpckhwd   mm4, mm0                ;mm4 = 00|0d|00|0c
      pmaddwd     mm2, mm3                ;mm2 =  b*co|a*co
      pmaddwd     mm4, mm3                ;mm4 =  d*co|c*co
      paddd       mm1, mm2
      paddd       mm7, mm4                ;accumulate
      add         eax, src_pitch          ;srcp2 += src_pitch
      inc         ebx
      cmp         ebx, edi
      jnz         bloop
      mov         eax, dstp
      pslld       mm1, 2
      pslld       mm7, 2                  ;compensate the fact that FPScale = 16384
      packuswb    mm1, mm7                ;mm1 = d|_|c|_|b|_|a|_
      psrlw       mm1, 8                  ;mm1 = 0|d|0|c|0|b|0|a
      packuswb    mm1, mm2                ;mm1 = 0|0|0|0|d|c|b|a
      movd        [eax+ecx*4], mm1
      inc         ecx
      cmp         ecx, xloops
      jnz         xloop
      add         eax, dst_pitch
      lea         edx, [edx+edi*4]        ;cur += fir_filter_size
      mov         dstp, eax
      dec         y
      jnz         yloop
      emms
    }
};

YUY2_Frame temp_frame;

int	i_width = 0, i_height = 0, o_width = 0, o_height = 0;
double	o_b = -1, o_c = -1;

void Close_YUY2_Resize( void )
{
	i_width = i_height = o_width = o_height = 0;
	o_b = o_c = -1;

	aligned_free( V_resampling_pattern );
	V_resampling_pattern = 0;
	aligned_free( V_yOfs );
	V_yOfs = 0;
	aligned_free( H_pattern_luma );
	H_pattern_luma = 0;
	aligned_free( H_pattern_chroma );
	H_pattern_chroma = 0;
	aligned_free( H_tempUV );
	H_tempUV = 0;
	aligned_free( H_tempY );
	H_tempY = 0;
	aligned_free( temp_frame.data );
	temp_frame.data = 0;
}


void Run_YUY2_Resize( YUY2_Frame *src, YUY2_Frame *dst, bool flip, double b, double c )
{
	if ( src->width != i_width || src->height != i_height ||
		dst->width != o_width || dst->height != o_height ||
		o_b != b || o_c != c ) {
		i_width = src->width;
		i_height = src->height;
		o_width = dst->width;
		o_height = dst->height;
		o_b = b;
		o_c = c;

		aligned_free( V_resampling_pattern );
		V_resampling_pattern = 0;
		aligned_free( V_yOfs );
		V_yOfs = 0;
		aligned_free( H_pattern_luma );
		H_pattern_luma = 0;
		aligned_free( H_pattern_chroma );
		H_pattern_chroma = 0;
		aligned_free( H_tempUV );
		H_tempUV = 0;
		aligned_free( H_tempY );
		H_tempY = 0;
		aligned_free( temp_frame.data );
		temp_frame.data = 0;

		if ( i_width != o_width ) {
			H_tempY = (unsigned char*)aligned_malloc( o_width*2+4, 64);   // aligned for Athlon cache line
			H_tempUV = (unsigned char*)aligned_malloc( o_width*4+8, 64);  // aligned for Athlon cache line
			H_pattern_chroma = GetResamplingPatternYUV( i_width>>1, 0, i_width/2, o_width>>1, &MitchellNetravaliFilter( b, c ), false, H_tempUV);
			H_pattern_luma = GetResamplingPatternYUV( i_width, 0, i_width, o_width, &MitchellNetravaliFilter( b, c ), true, H_tempY);
		}

		if ( i_height != o_height ) {
			V_resampling_pattern = GetResamplingPatternRGB( i_height, 0, i_height, o_height, &MitchellNetravaliFilter( b, c ) );
		    V_yOfs = (int *)aligned_malloc( i_height*4+4, 64);
			for (int i = 0; i < i_height; i++) V_yOfs[i] = o_width * 2 * i;
		}

		temp_frame.data = 0;

		if ( i_width != o_width && i_height != o_height ) {
			temp_frame.data = (char*)aligned_malloc( o_width*i_height*2+4, 64 );
		}
	}
	
	const double subrange_left = 0, subrange_top = 0;
	const double subrange_width = src->width, subrange_height = src->height;
	bool H = ((subrange_width != src->width) || (dst->width != src->width));
	bool V = ((subrange_height != src->height) || (dst->height != src->height));
	// ensure that the intermediate area is maximal
//	const double area_FirstH = subrange_height * dst->width;
//	const double area_FirstV = subrange_width * dst->height;

//	if ( area_FirstH < area_FirstV )
//	{
/*
		if ( V && H ) {
			temp_frame.data = (char*)aligned_malloc( src->width*dst->height*2+4, 64 );
			temp_frame.width = src->width;
			temp_frame.height = dst->height;
			temp_frame.pitch = src->width*2;
			temp_frame.rowsize = src->width*2;
			FilteredResizeV( src, &temp_frame, subrange_top, subrange_height, &MitchellNetravaliFilter( b, c ) );
			FilteredResizeH( &temp_frame, dst, subrange_left, subrange_width, &MitchellNetravaliFilter( b, c ) );
			aligned_free( temp_frame.data );
		} else {
			if (V)
				FilteredResizeV( src, dst, subrange_top, subrange_height, &MitchellNetravaliFilter( b, c ) );
			if (H)
				FilteredResizeH( src, dst, subrange_left, subrange_width, &MitchellNetravaliFilter( b, c ) );
		}
	} else {
*/
		if ( V && H ) {
			temp_frame.width = dst->width;
			temp_frame.height = src->height;
			temp_frame.pitch = dst->width*2;
			temp_frame.rowsize = dst->width*2;
			FilteredResizeH( src, &temp_frame, subrange_left, subrange_width, 0 );
			FilteredResizeV( &temp_frame, dst, subrange_top, subrange_height, flip );
		} else {
			if (H)
				FilteredResizeH( src, dst, subrange_left, subrange_width, flip );
			if (V)
				FilteredResizeV( src, dst, subrange_top, subrange_height, flip );
		}

		if ( !V && !H ) {
			int	x_count = src->width/4;
			char *s_data = src->data;
			char *d_data = dst->data;
			int s_pitch = src->pitch;
			int d_pitch = dst->pitch;
			int	s_height = src->height;
			if ( flip ) {
				d_data = d_data+d_pitch*(dst->height-1);
				d_pitch = -d_pitch;
			}
__asm {
		mov			eax,s_data;
		mov			ecx,d_data;
		mov			ebx,s_height

loopy:	mov			esi,eax
		mov			edi,ecx
		mov			edx,x_count
loopx:	movq		mm7,[esi]
		add			esi,8
		movq        [edi],mm7
		add			edi,8
		dec			edx
		jnz			loopx

		add			eax,s_pitch
		add			ecx,d_pitch

		dec			ebx
		jnz			loopy

overb:	emms
		}
	}

//	}
}
